# version 2022.1
[CmdLetBinding()]
param(
	[string]$username,
	[string]$password,
	[Switch]$logToConsole,
	[string]$inputFile
)

try {
	cd $PSScriptRoot
	$logFileName = 'workflow_events.{0}.log' -f (get-date -Format 'yyyy-MM')
	. ./utils.ps1 -logFileName $logFileName -logToConsole:$logToConsole
	$api = [mm]::getAPI($username, $password)
} catch {
	Write-Output ('Got exception while initializing: ' + $_.Exception.GetBaseException().Message)
	exit 1
}

. ./workflow_common.ps1


Enum Templates {
	processed
	rejected
	failed
	submitted
	approved
}

function sendEmail([string[]]$to, [Templates]$template, [hashtable]$extraTokens = $null, [switch]$includeState, $cc = $null){
	function getRequestDescription (){
		$changes = @()
		if ($objType -eq 'DNSRecord'){
			foreach ($change in $changeRequest.dnsRecordChanges.dnsRecordChange){
				$changes += $change
			}
			return [htmlUtils]::getRecordChangeListAsHTML($changes,$includeState)
		} else {
			throw "Don't know how to get request description for $objType requests"
		}
	}

	try {
		$tokens = @{
			requester = $requester
			requestID = $requestID
			requestType = $requestType
			application = $requestPropMap.Application
			group = (?? $extraTokens.Team $requestPropMap.Group) # should be TeamDL on records in Predefined approval group requests, but Group is the TeamDL in peer review requests
			requestComment = $changeRequest.saveComment
			scheduledDate = (?? $changeRequest.requestDate 'Immediately after approval')
			description = getRequestDescription
		}
		if ($null -ne $extraTokens){
			$tokens += $extraTokens
		}

		switch ($template){
			processed {
				$file    = "request-processed.htm"
				$subject = "Your Men&Mice $requestType Request has been implemented"
			}
			rejected {
				$file    = "request-rejected.htm"
				$subject = "Your Men&Mice $requestType Request has been rejected"
			}
			failed {
				$file    = "request-failed.htm"
				$subject = "Error(s) occurred while implementing Men&Mice $requestType Request #$requestID"
			}
			submitted {
				$file    = "request-submitted.htm"
				$subject = "A new Men&Mice $requestType Request is pending your review"
			}
			approved {
				$file    = "scheduled-request-approved.htm"
				$subject = "Your scheduled Men&Mice $requestType Request has been approved"
			}
		}

		if ($file -and $subject){
			[mmEmail]::prepareAndSend($to, $subject, $file, $cc, $mmConfig.email.bcc, $tokens)
		} else {
			log 'Unexpected Error: No template file or subject'
		}
	} catch {
		logException 'while sending email'
	}
}

function isFullyApproved () {
	$stateMap = @{}
	foreach ($change in [mm]::getFlatChangeListFromRequest($changeRequest)){
		if (notEmpty $change){
			$stateMap[$change.state] = $true
		}
	}

	foreach ($state in $stateMap.Keys){
		if ((notEmpty $state) -and $state -ne 'Approved'){
			logv2 ('Request is not fully approved')
			return $false
		}
	}

	logv2('Request is fully approved')
	return $true
}

function applyDNSRecordRequest () {

	function applyRecordProperties() {
		foreach ($change in [mm]::getFlatChangeListFromRequest($changeRequest)){
			if ($change.state -ne 'Applied'){
				# don't want to process changes that aren't applied, likely failed
				# likely all of them are failed - so don't log anything
				continue
			}

			$rec = $change.dnsRecordAfter
			if (!$rec){
				# only process modify or add
				continue
			}

			$recProps = [mm]::getPropertyMap($rec.customProperties)

			function setProp($name, $value){
				# so that email will include these added properties
				if ($rec.customProperties.property -is [array]){
					foreach ($property in $rec.customProperties.property){
						if ($property.name -eq $name){
							$property.value = $value
							return
						}
					}
					$rec.customProperties.property += @{name=$name; value=$value}
				}
			}
			
			try {
				$recordPropsToSet = @{
					Date = [mm]::getCurrentDay()
					Request = $requestID
					State = 'OK' # always set to OK. For zones that don't have attestation, will just always be OK
				}
				setProp 'Date' $recordPropsToSet.Date
				# don't set Request, because we don't want/need that in the summary
				setProp 'State' $recordPropsToSet.State
				
				$manager = [adUtils]::getManagerForUser($recProps.Owner).name
				if ([mm]::notEmpty($manager)){
					if ($manager -ne $recProps.Manager){
						log "Manager for $($recProps.Owner) has changed: $($recProps.Manager) -> $manager"
						$recordPropsToSet.Manager = $manager
						setProp 'Manager' $manager
					}
				} else {
					log "Manager not found in AD for Owner `"$($recProps.Owner)`""
				}

				if ($requestPropMap -eq $kWorkflowMap.peer){
					# we want to copy the Group from the request if the record is being created/modified to end up with no TeamDL
					# We did verify when submitting that there is only a single Team on the changes, and that it's consistent with the Group on the request
					# so we can simply blindly set the TeamDL to Group here - nothing will be changed if already the correct value
					$recordPropsToSet.TeamDL = $requestPropMap.Group
					setProp 'TeamDL' $recordPropsToSet.TeamDL
				}
				
				log ("Setting ownership properties for $($rec.ref): " + [mmJSON]::stringify($recordPropsToSet))
				$api.setProperties($rec.ref, 'DNSRecord', $recordPropsToSet, $IDsaveComment)
			} catch {
				logException "while setting properties for record $($rec.name)"
			}
		}
	}

	function sendProcessedEmail(){
		$requesterEmail = [adUtils]::getEmailForUser($requester)

		# collect emails for all owners and teams on all changes, before and after
		$ownersAndTeams = @{}
		foreach ($change in [mm]::getFlatChangeListFromRequest($changeRequest)){
			$props1 = [mm]::getPropertyMap($change.dnsRecordBefore.customProperties) 
			$props2 = [mm]::getPropertyMap($change.dnsRecordAfter.customProperties)


			if ([mm]::notEmpty($props1.Owner) -and $props1.Owner -ne $requester -and !$ownersAndTeams.ContainsKey($props1.Owner)){
				$ownersAndTeams[$props1.Owner] = [adUtils]::getEmailForUser($props1.Owner)
			}
			if ([mm]::notEmpty($props2.Owner) -and $props2.Owner -ne $requester -and !$ownersAndTeams.ContainsKey($props2.Owner)){
				$ownersAndTeams[$props2.Owner] = [adUtils]::getEmailForUser($props2.Owner)
			}

			if ([mm]::notEmpty($props1.TeamDL) -and !$ownersAndTeams.ContainsKey($props1.TeamDL)){
				$ownersAndTeams[$props1.TeamDL] = [adUtils]::getGroup($props1.TeamDL).mail
			}
			if ([mm]::notEmpty($props2.TeamDL) -and !$ownersAndTeams.ContainsKey($props2.TeamDL)){
				$ownersAndTeams[$props2.TeamDL] = [adUtils]::getGroup($props2.TeamDL).mail
			}
		}

		# we do want to send all record owners and teams involved in the request telling them it has been approved and applied
		# the requester is still in the to address, but any owner (if different from requester) and team that any change in the request is relevant for gets informed via cc,
		# whether it be a new Team a record is being assigned to or the old or new owner if being changed, etc
		log "Sending processed email notification to requester, with record owners and teamsDLs in cc"
	
		$cc = [string[]]( $ownersAndTeams.Values | Where-Object { [mm]::notEmpty($_) }) -join ';'
		sendEmail -to @($requesterEmail) -cc $cc -template ([Templates]::processed) -includeState
	}

	function sendFailureEmail(){

		# start by getting the failure errors
		try {
			$events = $api.call('GetHistory', @{objRef = $requestRef; startDate = $theEvent.date; endDate = $theEvent.date; eventType = 'Error'}).events
		} catch {
			logException 'while retrieving error history'
		}

		$theErrors = @()
		foreach ($ev in $events){
			$theErrors += ($ev.eventText -replace '\[Request ID\: \d+\]$','').replace('Got an error while applying the change request:','')
		}

		if (!$theErrors.Count) {
			log ('Unable to find any appropriate Failure/Error messages. Using "Unknown error. Please inform dns@intel.com"')
			$theErrors = @('Unknown error.')
		} else {
			log "The request failed with error: \n\t$($theErrors -join '\n\t')"
		}

		$extraTokens = @{
			reason = [htmlUtils]::getArrayAsHTMListItems($theErrors)
		}

		$group = [adUtils]::getGroup($requestPropMap.Group)

		if ($workflowType -eq $kWorkflowMap.peer){
			# send the failure notification to the requester and have the group on cc
			log "Sending failure email to requester $requester with team/group $($group.name) on cc"
			sendEmail -to @([adUtils]::getEmailForUser($requester)) -template ([Templates]::failed) -includeState -extraTokens $extraTokens -cc $group.mail
		}
		if ($workflowType -eq $kWorkflowMap.specificGroup){
			# here, we'll ask approval group to reapprove but add the requester on cc, because the requester might also be able to retry
			log "Sending failure email to approval group $($group.name) with requester $requester on cc"
			sendEmail -to @($group.mail)  -template ([Templates]::failed) -includeState -extraTokens $extraTokens -cc ([adUtils]::getEmailForUser($requester))
		}
	}

	# apply record properties regardless of state - will only apply to changes in applied state
	applyRecordProperties

	if ($changeRequest.state -eq 'Applied'){
		sendProcessedEmail
	} elseif ($changeRequest.state -eq 'Failed'){
		sendFailureEmail
	} else {
		log "ASSERT: Unexpected request state for an applied request $($changeRequest.state)"
	}
}

function applyCustomPropertyChanges(){
	if ($workflowType -eq $kWorkflowMap.renewal){
		foreach ($change in [mm]::getFlatChangeListFromRequest($changeRequest)){
			if ($change.state -ne 'Applied'){
				# don't want to process changes that aren't applied, likely failed
				# likely all of them are failed - so don't log anything
				continue
			}
			if (!($change.ref.startsWith('DNSRecords/') -or $change.ref.startsWith("{#13-#"))){
				log "ASSERT: not a DNS record change? $($change.ref)"
			}
			
			$before = [mm]::getPropertyMap($change.customPropertiesBefore)
			$after  = [mm]::getPropertyMap($change.customPropertiesAfter)
			if ($after.State -eq 'OK' -and $before.State -ne 'OK'){
				# state is being set to OK from something else ... we can safely just reset the date then too
				try {
					$api.setProperties($change.ref, 'DNSRecord', @{Date=[mm]::getCurrentDay()}, "Date reset when record was renewed.")
				} catch {
					logException "while setting properties for record $($rec.name)"
				}
			}	
		
		}		
	}
}

function processRejectChanges () {
	try {
		$tokens = @{
			reason = $theEvent.comment
			rejecter = if ($isScriptRunner) {(?? $mmConfig.email.scriptUserFriendlyName 'Micetro')} else { $invokee }
		}

		$manager = [adUtils]::getManagerForUser($requester)

		log "Sending a rejection notification to $requester, $($manager.name) on cc"
		sendEmail -to @([adUtils]::getEmailForUser($requester)) -cc $manager.mail -template ([Templates]::rejected) -extraTokens $tokens -includeState
	} catch {
		logException 'while sending rejection notification'
	}
}

function processApproveChanges () {

	if (isFullyApproved){
		log ('Request is now fully approved.')
		# If something needs to be done after approval but before the changes are applied, do it here

		if ($isScriptRunner) {
			log 'Detected that request is being automated by workflow'
		}

		if ($isScheduled){
			log "Sending notification to $requester that scheduled request has been approved"
			sendEmail -to @([adUtils]::getEmailForUser($requester)) -template ([Templates]::approved) -includeState
		}

	} else {
		log ('Request *not* fully approved yet.')
	}
}

function processAddChangeRequest () {
	function automateRecordOwnershipOnlyChanges() {
		# all changes that are modifications, where only the custom properties change, 
		# i.e. name, type, data, ttl stay the same between before and after
		# can be automated because we've already prevented unauthorized users (outside of ownership) to request changes to the record at all
		$changes = [mm]::getChangesFromRequest($changeRequest).dnsRecordChanges
		$propertyOnlyChanges = @()
		foreach ($change in $changes){
			if ($change.type -eq $kChangeType.mod -and [mm]::compareRecords($change.dnsRecordBefore, $change.dnsRecordAfter)){
				$propertyOnlyChanges += $change
			}
		}
		
		if ($propertyOnlyChanges.Count -eq 0){
			log 'None of the changes were ownership/property update'
			return $false
		}
	
		$cnt = $propertyOnlyChanges.Count
		
		$setting = ?? $Global:mmConfig.automateRecordOwnershipChanges 'always'
		if ($setting -eq 'always') {
			log "Automating $cnt ownership changes (always)"
		} elseif ($setting -eq $workflowType) {
			log "Automating $cnt ownership changes (matches workflow $workflowType)"
		} else {
			log "Ownership changes not automated. Set to '$setting', workflow is '$workflowType'"
			return $false
		}
	
		$saveComment = "Automated as part of workflow for Request #$requestID. Only Ownership/properties changed on record."
		# there's some mutex causing issues such that 
		$api.callAsync('ApproveChanges',@{changeRequestRef=$changeRequest.ref; dnsRecordChanges=$propertyOnlyChanges; saveComment=$saveComment})

		return $changes.Count -gt 0 -and $changes.Count -eq $propertyOnlyChanges.Count
	}

	function handlePredefinedApprovalRequest(){

		$groupInfo = [adUtils]::getGroup($requestPropMap.Group)
		if ($groupInfo -and $groupInfo.mail){
			log "Sending notification to approval group group email $($groupInfo.mail)"
			
			# since the email template has "Group" in it, that by default is set to the request group, which is the Approval group 
			# let's collect TeamDLs (most likely just a single one) from the records and replace "Group" with that here so that e.g. ISEC knows which Team is requesting this change
			$teamDLs = @{}
			foreach ($change in $changeRequest.dnsRecordChanges.dnsRecordChange){
				$propsBefore = [mm]::getPropertyMap($change.dnsRecordBefore.customProperties)
				$propsAfter = [mm]::getPropertyMap($change.dnsRecordAfter.customProperties)
				if ($propsBefore.TeamDL -and !$teamDLs.ContainsKey($propsBefore.TeamDL) -and [adUtils]::verifyGroup($propsBefore.TeamDL)){
					$teamDLs[$propsBefore.TeamDL] = $true
				}
				if ($propsAfter.TeamDL -and !$teamDLs.ContainsKey($propsAfter.TeamDL) -and [adUtils]::verifyGroup($propsAfter.TeamDL)){
					$teamDLs[$propsAfter.TeamDL] = $true
				}
			}

			$teamsStr = [string[]]@($teamDLs.Keys)
			log "Setting Group to TeamDL(s) found in records: $teamsStr"
			
			sendEmail -to @($groupInfo.mail) -template ([Templates]::submitted) -extraTokens @{Team = $teamsStr} -cc ([adUtils]::getEmailForUser($requester))
		} else {
			log "Approval group not found in AD or mail not configured. name: $($groupInfo.name), mail: $($groupInfo.mail)" # both empty when not found ...
		}
	}

	function handlePeerReviewRequest(){
		$manager = [adUtils]::getManagerForUser($requester)
		$groupInfo = [adUtils]::getGroup($requestPropMap.Group)
		if ($groupInfo -and $groupInfo.mail){
			log "Sending notification to group $($groupInfo.mail), $($manager.mail) on cc"
			sendEmail -to @($groupInfo.mail) -template ([Templates]::submitted) -cc $manager.mail
		} else {
			log ('ASSERT: Group {0} not found or mail not configured. Sending to manager {0} only.' -f $requestPropMap.Group, $manager.mail)
			sendEmail -to @($manager.mail) -template ([Templates]::submitted)
		}
	}

	# when request has been fully automated, we don't want to do anything here - we just notify when applied
	# here we only want to notify the approval group that a request is pending their review
	# we wont inform a team/owner that a record is _being_ assigned to them here ... we'll just inform them about that once it has happened

	if (isFullyApproved){
		log "Not sending any pending approval notification because all of the changes have already been approved"
		return
	}

	try {
		if ($workflowType -eq $kWorkflowMap.peer){
			handlePeerReviewRequest
		}
		if ($workflowType -eq $kWorkflowMap.specificGroup){
			handlePredefinedApprovalRequest
		}
		if ($workflowType -eq $kWorkflowMap.ownership){
			$didAutomate = automateRecordOwnershipOnlyChanges
			if (!$didAutomate){
				log "Ownership request not automated. Handling as peer approval request for notification and approval"
				handlePeerReviewRequest
			}
		}
	} catch {
		logException 'while processing AddChangeRequest'
	}
}

function processCheckApproveRejectChanges(){
	$orgAction = $eventAction.replace('ChangesCheck','') # => Approve or Reject
	if ($changeRequest.state -eq 'Failed') {
		log "Request is in Failed state. Anyone allowed to $orgAction request, to either retry or give up on request - allowed"
		return
	}

	# if user has access to all the changes, we assume it's an admin and simply allow it
	# if not, we check that the invokee is a member of the group registered on the request

	function hasAccessToAllPendingChanges () {
		foreach ($change in [mm]::getFlatChangeListFromRequest($changeRequest)){
			if ($change.state -eq 'Pending' -and [int]($change.hasAccess) -eq 0){
				return $false
			}
		}
		return $true
	}

	if (hasAccessToAllPendingChanges){
		if ($invokee -eq 'ScriptRunner'){
			log 'ScriptRunner is automating changes - always allowed'
			return
		}
		if ($workflowType -eq $kWorkflowMap.peer){
			log "$invokee has access to all changes in the request. Group membership not checked - allowed"
			return
		}
		if ($workflowType -eq $kWorkflowMap.specificGroup){
			log "$invokee has access to all changes in the request. By policy, only Approval Group members can approve a pending request. Verifying if member..."
		}
	}

	$approvalGroup = $requestPropMap.Group

	if ($invokee -eq $requester){
		# even a member of the security group cannot approve his own requests in specificGroup requests
		$msg = "You cannot $orgAction your own request"
		log "Returning error: $msg"
		[Console]::Out.Write($msg)
		exit 1
	} elseif ([adUtils]::isUserMemberInGroup($invokee, $approvalGroup)){
		log "$invokee is a member in approval group $approvalGroup - allowed"
	} elseif ($workflowType -eq $kWorkflowMap.peer -and $invokee -eq [adUtils]::getManagerForUser($requester).name){
		# only in peer review requests the manager can approve
		log "$invokee is the manager for requester $requester - allowed"
	} else {
		$msg = "You need to be a member of $approvalGroup to $orgAction this request"
		log "Returning error: $msg"
		[Console]::Out.Write($msg)
		exit 1
	}
}



try {
if ([mm]::notEmpty($inputFile)){
	logv2 "Reading input from $inputFile"
	$jsonStr = Get-Content -Raw $inputFile
} else {
	$jsonStr = [Console]::In.ReadToEnd()
	logv2 $jsonStr
}

$jsonInput = [mmJSON]::parse($jsonStr)
$inputParams = $jsonInput.externalScriptParameters
$theEvent = $inputParams.event
$eventAction = $theEvent.action
if ([mm]::isEmpty($inputFile)){
	[mm]::storeDebugBlob($jsonStr, $eventAction, 'json')
}
$changeRequest = ?? $inputParams.requestAfter.changeRequest $inputParams.requestBefore.changeRequest
$requestPropMap = [mm]::getPropertyMap($changeRequest.customProperties)
$objType = $changeRequest.objType
$workflowType = $requestPropMap.Workflow
$requestRef = $changeRequest.ref
$requestID = [mm]::ref2ID($requestRef)
$requestType = 'DNS'

$requester = [mm]::cleanXmlString($changeRequest.requester)
$invokee = [mm]::cleanXmlString($theEvent.userName)
$isScriptRunner = $invokee -eq 'ScriptRunner'
$isScheduled = [mm]::notEmpty($changeRequest.requestDate)
$IDsaveComment = "RequestID:$requestID"


log ('Processing #{0}: {1}. Requested by: {2}. Command invoked by: {3}' -f $requestID,$eventAction,$requester,$invokee)

try {
	switch ($eventAction) {
		'applyDNSRecordChanges' { applyDNSRecordRequest ; break}
		'applyCustomPropertyChanges' { applyCustomPropertyChanges ; break}
		'RejectChanges' { processRejectChanges ; break }
		'RejectDNSRecordChanges' { processRejectChanges ; break }
		'ApproveChanges' { processApproveChanges ; break }
		'ApproveDNSRecordChanges' { processApproveChanges ; break }
		'AddChangeRequest' { processAddChangeRequest ; break }
		'AddDNSRecordChangeRequest' { processAddChangeRequest ; break }
		'ApproveChangesCheck' { processCheckApproveRejectChanges; break }
		'RejectChangesCheck' { processCheckApproveRejectChanges; break }
		default {
			log ('ERROR: Unsupported eventAction "{0}" ' -f $eventAction)
			exit 1
		}
	}
} catch {
	logException 'while processing request event. Notify M&M'
}

log 'Done'

} catch {
	logException 'during processing'
} finally {
	$api.logout()
}
